import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import os
import random
import numpy as np
from collections import deque

#LOTE_MENOR = 5

CAMADA_ENTRADA = 22
CAMADA_SAIDA = 4


"""
    Classe base do modelo de rede neural.
"""
class Linear_QNet(nn.Module):
    def __init__(self, tamanho_entrada, tamanho_oculto, tamanho_saida, dispositivo, caminho_modelo):
        super().__init__()
        self.dispositivo = dispositivo
        self.linear1 = nn.Linear(tamanho_entrada, tamanho_oculto).to(self.dispositivo)
        self.linear2 = nn.Linear(tamanho_oculto, tamanho_saida).to(self.dispositivo)
        self.caminho_modelo = caminho_modelo
    
    # cria uma predição baseada na entrada
    def forward(self, x):
        x = F.relu(self.linear1(x))
        x = self.linear2(x)
        return x

    # salva o modelo
    def Save(self, nome_arq="model.pth"):
        if not os.path.exists(self.caminho_modelo):
            os.makedirs(self.caminho_modelo)
        
        nome_arq = os.path.join(self.caminho_modelo, nome_arq)
        torch.save(self.state_dict(), nome_arq)

    # carrega o modelo
    def Load(self, nome_arq):
        self.load_state_dict(torch.load(nome_arq, map_location=self.dispositivo))

"""
    Classe do agente usando o modelo de rede neural.
"""
class Agente:
    def __init__(self, eps_ini, gamma, taxa_aprendizado, camada_oculta, tamanho_memoria, tamanho_lote, tamanho_menor, usar_gpu, caminho_modelo):
        self.n_partidas = 0
        self.epsilon = eps_ini
        self.gamma = gamma
        self.tamanho_memoria = tamanho_memoria
        self.tamanho_lote = tamanho_lote
        self.tamanho_lote_menor = tamanho_menor
        self.dispositivo = torch.device("cuda" if torch.cuda.is_available() else "cpu") if usar_gpu == "True" else 'cpu'
        self.memoria = deque(maxlen=self.tamanho_memoria)
        self.modelo = Linear_QNet(CAMADA_ENTRADA, camada_oculta, CAMADA_SAIDA, self.dispositivo, caminho_modelo).to(self.dispositivo)
        self.modelo_alvo = Linear_QNet(CAMADA_ENTRADA, camada_oculta, CAMADA_SAIDA, self.dispositivo, caminho_modelo).to(self.dispositivo)
        self.modelo_alvo.load_state_dict(self.modelo.state_dict())
        self.lr = taxa_aprendizado
        self.optimizer = optim.Adam(self.modelo.parameters(), lr=self.lr)
        self.criterion = nn.SmoothL1Loss()

    # guarda as informações da ação executada
    def Guardar(self, estado, acao, recompensa, proximo_estado, concluido):
        self.memoria.append((estado, acao, recompensa, proximo_estado, concluido))


    def LimparMemoria(self):
        self.memoria.clear()

    # treina o modelo usando um lote de ações aleatórias da memória
    def TreinarMemoriaLonga(self):
        if len(self.memoria) > self.tamanho_lote:
            amostra = random.sample(self.memoria, self.tamanho_lote)
        else:
            amostra = self.memoria

        estados, acoes, recompensas, proximo_estados, concluidos = zip(*amostra)

        return self.Treinamento(estados, acoes, recompensas, proximo_estados, concluidos)
    
    # treina o modelo usando a última ação executada
    def TreinarMemoriaCurta(self):
        if len(self.memoria) > self.tamanho_lote_menor:
            amostra = random.sample(self.memoria, self.tamanho_lote_menor)
        else:
            amostra = self.memoria

        estados, acoes, recompensas, proximo_estados, concluidos = zip(*amostra)

        return self.Treinamento(estados, acoes, recompensas, proximo_estados, concluidos)
    

    # Gera uma ação usando o modelo baseada no estado atual ou uma ação aleatória
    def GerarAcao(self, estado):
        move = [0] * CAMADA_SAIDA
        aux = []
        if random.random() < self.epsilon:
            aux = (random.randint(0, CAMADA_SAIDA-1))
            move[aux] = 1
        else:
            estado_torch = torch.tensor(estado, dtype=torch.float, device=self.dispositivo)
            if self.dispositivo != 'cpu':
              predicao = self.modelo(estado_torch.cuda())
            else:
                predicao = self.modelo(estado_torch)
            aux = torch.argmax(predicao).item()
            move[aux] = 1

        return move

    # treina o modelo com os dados fornecidos
    def Treinamento(self, estado, acao, recompensa, proximo_estado, concluido):
        estado_np = np.array(estado)
        proximo_estado_np = np.array(proximo_estado)
        acao_np = np.array(acao)
        recompensa_np = np.array(recompensa)


        estado = torch.tensor(estado_np, dtype=torch.float, device=self.dispositivo)
        proximo_estado = torch.tensor(proximo_estado_np, dtype=torch.float, device=self.dispositivo)
        acao = torch.tensor(acao_np, dtype=torch.long, device=self.dispositivo)
        recompensa = torch.tensor(recompensa_np, dtype=torch.float, device=self.dispositivo)
        # (n, x)

        if len(estado.shape) == 1:
            # (1, x)
            estado = torch.unsqueeze(estado, 0)
            proximo_estado = torch.unsqueeze(proximo_estado, 0)
            acao = torch.unsqueeze(acao, 0)
            recompensa = torch.unsqueeze(recompensa, 0)
            concluido = (concluido, )

        pred = self.modelo(estado)

        alvo = pred.clone()
        for idx in range(len(concluido)):
            novo_Q = recompensa[idx]
            if not concluido[idx]:
                novo_Q = recompensa[idx] + self.gamma * torch.max(self.modelo_alvo(proximo_estado[idx]))

            alvo[idx][torch.argmax(acao[idx]).item()] = novo_Q
    
        # novo_Q = r + y * max(next_predicted Q value) -> only do this if not concluido
        self.optimizer.zero_grad()
        perda = self.criterion(alvo, pred)
        perda.backward()

        self.optimizer.step()

        return perda.item()
    
    def AtualizaAlvo(self):
        self.modelo_alvo.load_state_dict(self.modelo.state_dict())
